#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This work was created by participants in the DataONE project, and is
# jointly copyrighted by participating institutions in DataONE. For
# more information on DataONE, see our web site at http://dataone.org.
#
# Copyright 2009-2015 DataONE
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''':mod:`views.v2`
==================

:Synopsis: REST call handlers for v2 of the DataONE Member Node APIs.
:Author: DataONE (Flynn)
'''
# Stdlib.
import re
import logging

# App.
import mn.auth as auth
import mn.models as models
import mn.view_asserts as view_asserts
import mn.view_shared as view_shared
import mn.sysmeta_validate as sysmeta_validate
import mn.sysmeta_store as sysmeta_store
import mn.util as util
from . import v1
import d1_common.date_time
import settings
from rest_framework.decorators import api_view


@api_view(['GET', 'POST', 'HEAD', 'PUT', 'DELETE'])
def dispatch_object_sid(request, sid=None, pid=None):
    if sid:
        pid = convert_sid_to_pid(sid)
    elif pid:
        view_asserts.pid_not_obsoleted(pid)
    if request.method == 'GET':
        return v1.get_object_pid(request, pid)
    elif request.method == 'HEAD':
        return v1.head_object_pid(request, pid)
    elif request.method == 'PUT':
        return put_object_sid(request, pid, sid)
    elif request.method == 'DELETE':
        return v1.delete_object_pid(request, pid)
    else:
        return v1.HttpResponseNotAllowed(['GET', 'HEAD', 'POST', 'PUT', 'DELETE'])

@api_view(['PUT', 'GET'])
def dispatch_meta(request, pid):
    if request.method == 'PUT':
        return put_system_metadata(request, pid)
    if request.method == 'GET':
        return get_meta_sid(request, pid)


def convert_unicode_none(obj):
    if obj.obsoletes == u'None':
        obj.obsoletes = None
    if obj.obsoletedBy == u'None':
        obj.obsoletedBy = None
    return obj


def dispatch_object(request):
    if request.method == 'GET':
        return v1.get_object(request)
    elif request.method == 'POST':
        return object_post(request)
    else:
        return v1.HttpResponseNotAllowed(['GET', 'POST'])


def get_meta_sid(request, sid):
    pid = convert_sid_to_pid(sid)
    return v1.get_meta_pid(request, pid)


def get_log_sid(request):
    if 'sidFilter' in request.GET.keys():
        pid = convert_sid_to_pid(request.GET.get('sidFilter'))
        request.GET = request.GET.copy()
        request.GET['pidFilter'] = pid
    return v1.get_log(request)


def get_checksum_sid(request, sid):
    pid = convert_sid_to_pid(sid)
    return v1.get_checksum_pid(request, pid)


def get_replica_sid(request, sid):
    pid = convert_sid_to_pid(sid)
    return v1.get_replica_pid(request, pid)


def get_is_authorized_sid(request, sid):
    pid = convert_sid_to_pid(sid)
    return v1.get_is_authorized_pid(request, pid)


@auth.assert_write_permission  # OLD object
def put_system_metadata(request, pid):
    '''MNStorage.updateSystemMetadata(session, pid, sysmeta) → boolean
    '''
    try:
        if service.settings.REQUIRE_WHITELIST_FOR_UPDATE:
            auth.assert_create_update_delete_permission(request)
        view_asserts.pid_exists(pid)
        sciobj = models.ScienceObject.objects.get(pid=pid)
        sysmeta_saved = sysmeta_store.sysmeta(pid,sciobj.serial_version)
        old_sysmeta = sysmeta_saved.sysmeta_pyxb
        sysmeta_xml = request.data['sysmeta'].read().decode('utf-8')
        uploaded_sysmeta = view_shared.deserialize_system_metadata(
            sysmeta_xml.encode('utf-8'))
        new_sysmeta = old_sysmeta
        sysmeta_validate.update_sysmeta(uploaded_sysmeta, new_sysmeta)
        sysmeta_store.write_sysmeta_to_store(new_sysmeta)
        view_shared.update_sysmeta_db(new_sysmeta, pid)
        return view_shared.http_response_with_identifier_type(pid)
    except d1_common.types.exceptions.InvalidSystemMetadata, e:
        logging.error('Invalid system metadata:\n{0}'
                      .format(str(e)))
        raise e

def post_refresh_sid_system_metadata(request):
    '''MNStorage.systemMetadataChanged(session, pid, serialVersion,
                                     dateSysMetaLastModified) → boolean
    '''
    try:
        sid = request.POST.get('sid')
        view_asserts.post_has_mime_parts(request, (('field', 'sid'),
                                                   ('field', 'serialVersion'),
                                                   ('field', 'dateSysMetaLastModified'),
                                                   ))
        view_asserts.object_exists(request.POST.get('sid'))
        refresh_queue = models.SystemMetadataRefreshQueue()
        refresh_queue.object = models.ScienceObject.objects.get(pid=request.POST.get('sid'))
    except:
        pid = request.POST.get('pid')
        view_asserts.post_has_mime_parts(request, (('field', 'pid'),
                                                   ('field', 'serialVersion'),
                                                   ('field', 'dateSysMetaLastModified'),
                                                   ))
        view_asserts.object_exists(request.POST.get('pid'))
        refresh_queue = models.SystemMetadataRefreshQueue()
        refresh_queue.object = models.ScienceObject.objects.get(pid=request.POST.get('pid'))
    refresh_queue.serial_version = request.POST.get('serialVersion')
    refresh_queue.last_modified = d1_common.date_time.from_iso8601(
        request.POST.get('dateSysMetaLastModified'))
    refresh_queue.set_status('new')
    refresh_queue.save_unique()
    return view_shared.http_response_with_boolean_true_type()



def put_archive_sid(request, sid):
    pid = convert_sid_to_pid(sid)
    return v1.put_archive_pid(request, pid)


def _create(request, pid, sysmeta, sid):
    # view_asserts.xml_document_not_too_large(request._full_data.get('sysmeta'))
    view_asserts.xml_document_not_too_large(request.data['sysmeta'])

    view_asserts.obsoleted_by_not_specified(sysmeta)
    sysmeta_validate.validate_sysmeta_against_uploaded(request, pid, sysmeta)
    sysmeta_validate.update_sysmeta_with_mn_values(request, sysmeta)
    # d1_common.date_time.is_utc(sysmeta.dateSysMetadataModified)
    return view_shared.create(request, pid, sysmeta, sid)


@auth.decorator_assert_create_update_delete_permission
def object_post(request):
    '''MNStorage.create(session, pid, object, sysmeta) → Identifier
    '''
    view_asserts.post_has_mime_parts(request, (('field', 'pid'),
                                               ('field', 'sid'),
                                               ('file', 'object'),
                                               ('file', 'sysmeta')))
    sysmeta_xml = request.data['sysmeta'].read().decode('utf-8')
    sysmeta = view_shared.deserialize_system_metadata(
        sysmeta_xml.encode('utf-8'))
    view_asserts.obsoleted_by_not_specified(sysmeta)
    view_asserts.obsoletes_not_specified(sysmeta)
    new_pid = request.POST.get('pid')
    sid = request.POST.get('sid')
    _create(request, new_pid, sysmeta, sid)
    return view_shared.http_response_with_identifier_type(new_pid)

@auth.assert_write_permission  # OLD object
def put_object_sid(request, old_pid, sid):
    '''MNStorage.update(session, sid, object, newPid, sysmeta) → Identifier
  '''
    if service.settings.REQUIRE_WHITELIST_FOR_UPDATE:
        auth.assert_create_update_delete_permission(request)
    util.coerce_put_post(request)
    request.FILES['sysmeta'] = request.POST.get('sysmeta')
    request.FILES['object'] = request.POST.get('object')
    # file_obj = upload_put_files(request, request.data['sysmeta'])
    # view_asserts.post_has_mime_parts(request, (
    #     ('field', 'newPid'),
    #     ('file', 'object'),
    #     ('file', 'sysmeta'))
    #                                  )

    view_asserts.pid_exists(old_pid)
    view_asserts.pid_not_obsoleted(old_pid)
    view_asserts.sci_obj_is_not_replica(old_pid)
    # params = split_put_request_body(request)
    sysmeta_xml = request.data['sysmeta'].read().decode('utf-8')
    sysmeta = view_shared.deserialize_system_metadata(
        sysmeta_xml.encode('utf-8'))
    view_asserts.obsoletes_matches_pid_if_specified(sysmeta, old_pid)
    sysmeta.obsoletes = old_pid
    new_pid = request.data['newPid']
    _create(request, new_pid, sysmeta, sid)
    v1._set_obsoleted_by(old_pid, new_pid)
    return view_shared.http_response_with_identifier_type(new_pid)


def upload_put_files(request, file_name):
    upload_handlers = request.upload_handlers
    content_type   = str(request.META.get('CONTENT_TYPE', ""))
    content_length = int(request.META.get('CONTENT_LENGTH', 0))

    # if content_type == "":
    #     return HttpResponse(status=400)
    # if content_length == 0:
    #     # both returned 0
    #     return HttpResponse(status=400)

    content_type = content_type.split(";")[0].strip()
    try:
        charset = content_type.split(";")[1].strip()
    except IndexError:
        charset = "utf8"

    # we can get the file name via the path, we don't actually
    field_name = file_name
    counters = [0]*len(upload_handlers)

    for handler in upload_handlers:
        result = handler.handle_raw_input("",request.META,content_length,"","")

    for handler in upload_handlers:

        try:
            handler.new_file(field_name, file_name,
                             content_type, content_length, charset)
        except:
            break

    for i, handler in enumerate(upload_handlers):
        while True:
            chunk = request.read(handler.chunk_size)
            if chunk:

                handler.receive_data_chunk(chunk, counters[i])
                counters[i] += len(chunk)
            else:
                # no chunk
                break

    for i, handler in enumerate(upload_handlers):
        file_obj = handler.file_complete(counters[i])
        return file_obj
        # if not file_obj:
        #     # some indication this didn't work?
        #     return HttpResponse(status=500)
        # else:
        #     # handle file obj!


def split_put_request_body(request):
    params = request.body.split(',')
    regex = re.compile("")
    name = [i for i in params if not regex.search(i)]
    return name

def generate_sid(request):
    return v1.post_generate_identifier(request)


# Conversion of sids to pids
def convert_sid_to_pid(uid):
    '''
    :param uid: identifier of unkown type (sid or pid)
    :return: sid (if sid doesn't exist, otherwise it's a pid)
    '''

    list_ordered_obj_lists = []
    # # check to see if uid is a pid, if so return it
    # if models.ScienceObject.objects.filter(pid=uid, archived=False) \
    #         .exists():
    #     return uid

    if view_shared.test_object_exists(uid):
        return uid

    # get list of science objects with the same sid
    sci_objs = list(models.ScienceObject.objects.filter(sid=uid))
    if not sci_objs:
        raise ValueError("ERROR: no objects matching sid {0}".format(uid))
    new_sci_objs = []

    # convert u'None' to None
    for obj in sci_objs:
        new_sci_objs.append(convert_unicode_none(obj))

    ordered_obj_list, remaining_objs = order_pid_list(new_sci_objs)
    list_ordered_obj_lists.append(ordered_obj_list)
    list_ordered_obj_lists = add_remaining_objects(
        remaining_objs, list_ordered_obj_lists)
    final_obj_list, orphan_obj_list = connect_object_lists(
        list_ordered_obj_lists)

    # convert list of lists into a single list
    output_list = flatten_list(final_obj_list)

    if len(orphan_obj_list) > 0:
        # check if date modified of orphan objects (not connected by
        # obsoletes or obsoletedBy to other objects) has a 'mtime' later than the
        # last element of ordered list
        for orphan_obj in orphan_obj_list:
            output_list = [output_list]
            if orphan_obj[0].mtime >= output_list[-1][-1].mtime:
                output_list.append(orphan_obj)
        output_list = flatten_list(output_list)
    return output_list[-1].pid


def add_remaining_objects(remaining_objs, list_ordered_obj_lists):
    while remaining_objs:
        new_ordered_obj_list, remaining_objs = order_pid_list(remaining_objs)
        list_ordered_obj_lists.append(new_ordered_obj_list)
    return list_ordered_obj_lists


def connect_object_lists(list_ordered_obj_lists):
    # given lists of ordered object lists, connect head of older list to tail
    # of newer list
    final_obj_list = [list_ordered_obj_lists[0]]
    list_ordered_obj_lists.remove(list_ordered_obj_lists[0])
    orphan_obj_list = []
    while len(list_ordered_obj_lists) > 0:
        for obj_list in list_ordered_obj_lists:
            if final_obj_list[0][0].obsoletes == obj_list[-1].obsoletedBy \
                    and obj_list[-1].obsoletedBy is not None:
                final_obj_list.insert(0, obj_list)
                list_ordered_obj_lists.remove(obj_list)
            elif final_obj_list[-1][-1].obsoletedBy == obj_list[0].obsoletes \
                    and obj_list[0].obsoletes is not None:
                final_obj_list.append(obj_list)
                list_ordered_obj_lists.remove(obj_list)
            else:
                orphan_obj_list.append(obj_list)
                list_ordered_obj_lists.remove(obj_list)

    return final_obj_list, orphan_obj_list


def order_pid_list(sci_objs):
    # starting with the first object in list, traverse obsoletes and
    # obsoletedBy chains to put pids into proper order
    new_obj = sci_objs[0]
    sorted_objs = [new_obj]
    sci_objs.remove(new_obj)
    while sorted_objs[0].obsoletes is not None and new_obj is not None:
        new_obj = get_obj_by_pid(sci_objs, sorted_objs[0].obsoletes)
        if new_obj is not None:
            sorted_objs.insert(0, new_obj)
            sci_objs.remove(new_obj)
    new_obj = sorted_objs[-1]
    while sorted_objs[-1].obsoletedBy is not None and new_obj is not None:
        new_obj = get_obj_by_pid(sci_objs, sorted_objs[-1].obsoletedBy)
        if new_obj is not None:
            sorted_objs.append(new_obj)
            sci_objs.remove(new_obj)
    return sorted_objs, sci_objs


def get_obj_by_pid(sci_objs, pid):
    # get object referred to by pid
    for obj in sci_objs:
        if obj.pid == pid:
            return obj
    return None


def flatten_list(input_list):
    # convert list of lists to single list
    output_list = [item for sublist in input_list for item in sublist]
    return output_list
